package ca.recrown.islandsurvivalcraft.caching;

import java.util.HashMap;
import java.util.PriorityQueue;

public class Cache<Key, Value> {
    private final int maxSize;
    private final float initialUsageFactor;
    private final HashMap<Key, CacheDataSet<Key, Value>> data;
    private final PriorityQueue<UsageTracker<Key>> usageTrackers;

    public Cache(int maxSize, float initialUsageFactor) {
        data = new HashMap<>(maxSize);
        usageTrackers = new PriorityQueue<>(maxSize);
        this.maxSize = maxSize;
        this.initialUsageFactor = initialUsageFactor;
    }

    public Cache(int maxSize) {
        this(maxSize, 1f);
    }

    public Cache(float initialUsageFactor) {
        this(1024, initialUsageFactor);
    }

    public Cache() {
        this(1024);
    }

    public void setValue(Key key, Value value) {
        CacheDataSet<Key, Value> dataSet = null;
        if (data.containsKey(key)) {
            dataSet =  data.get(key);
            usageTrackers.remove(dataSet.getUsageTracker());
        } else {
            if (usageTrackers.size() >= maxSize) {
                data.remove(usageTrackers.poll().getKey());
            }
            int currentLowest = 0;
            if (!usageTrackers.isEmpty()) {
                currentLowest = usageTrackers.peek().getUsage();
            }
            dataSet = new CacheDataSet<>(new UsageTracker<>(key, (int) initialUsageFactor * currentLowest));
            data.put(key, dataSet);
        }
        dataSet.getUsageTracker().increaseUsage();
        usageTrackers.add(dataSet.getUsageTracker());
        dataSet.setValue(value);
    }


    /**
     * Retrieves cached value associated to the key.
     * Uses equals.
     * If does not exist, returns null.
     * @param key the key associated with the value.
     * @return the value associated to the key.
     */
    public Value getValue(Key key) {
        CacheDataSet<Key, Value> dataSet = null;
        dataSet = data.get(key);
        if (dataSet == null) return null;

        usageTrackers.remove(dataSet.getUsageTracker());
        dataSet.getUsageTracker().increaseUsage();
        usageTrackers.add(dataSet.getUsageTracker());
        return dataSet.getValue();
    }

    public boolean hasValue(Key key) {
        if (data.containsKey(key)) {
            return data.get(key).getValue() != null;
        }
        return false;
    }

    public void clearCache() {
        data.clear();
        usageTrackers.clear();
    }
}